from ngsolve import * # sloppy

from netgen.geom2d import SplineGeometry
geo = SplineGeometry()

pnts =[ (-12,-5), (-7,-5), (-5,-5), (-3,-5), (-3,-3), 
        (-3,-1), (-1,-1), ( 0,-1), ( 1,-1), ( 3,-1), 
        ( 3,-3), ( 3,-5), ( 5,-5), ( 7,-5), ( 12,-5), 
        ( 12, 5), ( 7, 5), ( 5, 5), ( 3, 5), ( 3, 3), 
        ( 3, 1), ( 1, 1), ( 0, 1), (-1, 1), (-3, 1), 
        (-3, 3), (-3, 5), (-5, 5), (-7, 5), (-12, 5)]
pnts = [geo.AppendPoint(*pnt) for pnt in pnts]

curves = [[["line",0,1],"wall",1, 0], [["line",1,2],"wall",1, 0],
          [["spline3",2,3,4],"wall",1, 0],[["spline3",4,5,6],"wall",1, 0],[["line",6,7],"wall",1, 0],
          [["line",7,22],"dam",1, 2],     # <--- dam interface
          [["line",7,8],"wall",2, 0],[["spline3",8,9,10],"wall",2, 0],
          [["spline3",10,11,12],"wall",2, 0],[["line",12,13],"wall",2, 0],[["line",13,14],"wall",2, 0],
          [["line",14,15],"wall",2, 0], # <--- right boundary
          [["line",15,16],"wall",2, 0],[["line",16,17],"wall",2, 0],
          [["spline3",17,18,19],"wall",2, 0],[["spline3",19,20,21],"wall",2, 0],
          [["line",21,22],"wall",2, 0],[["line",22,23],"wall",1, 0],
          [["spline3",23,24,25],"wall",1, 0],[["spline3",25,26,27],"wall",1, 0],
          [["line",27,28],"wall",1, 0],[["line",28,29],"wall",1, 0],
          [["line",29,0],"wall",1, 0]]   # <--- left boundary

for c,bc,l,r in curves:
    geo.Append(c,bc=bc,leftdomain=l, rightdomain=r) 
geo.SetMaterial(1,"upperlevel")
geo.SetMaterial(2,"lowerlevel")    

mesh = Mesh(geo.GenerateMesh(maxh=2))
mesh.Curve(3)
Draw(mesh)


true_compile=True

# ### Approximation space
order = 2
fes = L2(mesh,order=order,dim=3)


# ### initial and boundary conditions
U,V = fes.TnT() # "Trial" and "Test" function
h, hu, hv = U

# initial conditions
h0mat = {"upperlevel" : 10, "lowerlevel" : 2}
U0 = CoefficientFunction((CoefficientFunction([h0mat[mat] for mat in mesh.GetMaterials()]),0,0))

# boundary conditions
hbndreg = CoefficientFunction([{"wall" : h, "dam" : 0}[rg] for rg in mesh.GetBoundaries()])
hubndreg = CoefficientFunction([{"wall" : -hu, "dam" : 0}[rg] for rg in mesh.GetBoundaries()])
hvbndreg = CoefficientFunction([{"wall" : -hv, "dam" : 0}[rg] for rg in mesh.GetBoundaries()])

Ubnd = CoefficientFunction((hbndreg,hubndreg,hvbndreg))

# constant for gravitational force
g=1


# ### flux formulation and numerical flux choice (Lax-Friedrich)
n = specialcf.normal(mesh.dim)

def Max(u,v):
    return IfPos(u-v,u,v)

def F(U):
    h, hvx, hvy = U
    vx = hvx/h
    vy = hvy/h
    return CoefficientFunction(((hvx,hvy),
                                (hvx*vx + 0.5*g*h**2, hvx*vy),
                                (hvx*vy, hvy*vy + 0.5*g*h**2)),dims=(3,2))
def Fmax(A,B):
    ha, hua, hva = A
    hb, hub, hvb = B
    vnorma = sqrt(hua**2+hva**2)/ha
    vnormb = sqrt(hub**2+hvb**2)/hb
    return Max(vnorma+sqrt(g*A[0]),vnormb+sqrt(g*B[0]))

def Fhatn(U):
    Uhat = U.Other(bnd=Ubnd)
    return (0.5*F(U)+0.5*F(Uhat))*n + Fmax(U,Uhat)/2*(U-Uhat)


# ### DG formulation
def DGBilinearForm(fes,F,Fhatn,Ubnd):
    a = BilinearForm(fes, nonassemble=True)
    if true_compile:
        print("start compile (I):", end="")
    a += SymbolicBFI ( (- InnerProduct(F(U),grad(V).trans)).Compile(true_compile,wait=True))
    a += SymbolicBFI ( (InnerProduct(Fhatn(U),V)).Compile(true_compile,wait=True), element_boundary=True)
    if true_compile:
        print(" done.")
    return a

a = DGBilinearForm(fes,F,Fhatn,Ubnd)

from DGdiffusion import AddArtificialDiffusion

artvisc = Parameter(1.0)
if order > 0:
    if true_compile:
        print("start compile (II):", end="")
    AddArtificialDiffusion(a,Ubnd,artvisc)
    if true_compile:
        print(" done.")

gfu = GridFunction(fes)
gfh, gfhu, gfhv = gfu
gfvu = gfhu/gfh
gfvv = gfhv/gfh
momentum = CoefficientFunction((gfhu,gfhv))
velocity = CoefficientFunction((gfvu,gfvv))
gfu.Set(U0)
Draw(momentum,mesh,"mom")
Draw(velocity,mesh,"vel")
Draw(gfh,mesh,"h")


# ### Implicit Euler time stepping
def TimeLoop(a,gfu,dt,T,nsamplings=100):
    res = gfu.vec.CreateVector()
    fes = a.space
    t = 0
    nsteps = int(ceil(T/dt))    
    i = 0
    with TaskManager():
        while t <= T - 0.5*dt:
            a.Apply(gfu.vec, res)
            fes.SolveM(rho=CoefficientFunction(1),vec=res)
            gfu.vec.data -= dt * res
            t += dt
            if (i+1) % int(nsteps/nsamplings) == 0:
                Redraw()
            i+=1
            print("\rt=",t,end="")
        Redraw()     
TimeLoop(a,gfu,dt=0.0004,T=5)
